@@ -1,21 +1,18 @@
 from __future__ import annotations
 
 from collections.abc import Iterable, Iterator, Sequence
-from typing import Any, Callable, ClassVar, Generic, Protocol, Tuple, cast
+from typing import Any, Callable, ClassVar, Generic, Protocol, TypeVar
 from urllib.parse import unquote, urldefrag, urljoin
 
 from attrs import evolve, field
-from pyrsistent import PMap as PMapType, plist, pmap, pset
-from pyrsistent.typing import PList, PMap, PSet
+from rpds import HashTrieMap, HashTrieSet, List
 
 from referencing import exceptions
 from referencing._attrs import frozen
-from referencing.typing import URI, Anchor as AnchorType, D, Mapping
+from referencing.typing import URI, Anchor as AnchorType, D, Mapping, Retrieve
 
-EMPTY_RESOURCES: PMap[URI, Resource[Any]] = pmap({}, pre_size=64)
-EMPTY_ANCHORS = cast(PMap[Tuple[URI, str], AnchorType[Any]], EMPTY_RESOURCES)
-EMPTY_UNCRAWLED: PSet[URI] = pset(pre_size=128)
-EMPTY_PREVIOUS_RESOLVERS: PList[URI] = plist()
+EMPTY_UNCRAWLED: HashTrieSet[URI] = HashTrieSet()
+EMPTY_PREVIOUS_RESOLVERS: List[URI] = List()
 
 
 class _MaybeInSubresource(Protocol[D]):
@@ -203,14 +200,6 @@ def _fail_to_retrieve(uri: URI):
     raise exceptions.NoSuchResource(ref=uri)
 
 
-def _to_pmap(
-    value: dict[URI, Resource[D]]
-    | PMap[URI, Resource[D]]
-    | list[tuple[URI, Resource[D]]],
-):
-    return value if isinstance(value, PMapType) else pmap(value)
-
-
 @frozen
 class Registry(Mapping[URI, Resource[D]]):
     r"""
@@ -236,15 +225,13 @@ class Registry(Mapping[URI, Resource[D]]):
     even according to the retrieval logic.
     """
 
-    _resources: PMap[URI, Resource[D]] = field(
-        default=EMPTY_RESOURCES,
-        converter=_to_pmap,
+    _resources: HashTrieMap[URI, Resource[D]] = field(  # type: ignore[reportGeneralTypeIssues]  # noqa: E501
+        default=HashTrieMap(),
+        converter=HashTrieMap.convert,
     )
-    _anchors: PMap[tuple[URI, str], AnchorType[D]] = field(
-        default=EMPTY_ANCHORS,
-    )
-    _uncrawled: PSet[URI] = field(default=EMPTY_UNCRAWLED)
-    _retrieve: Callable[[URI], Resource[D]] = field(default=_fail_to_retrieve)
+    _anchors: HashTrieMap[tuple[URI, str], AnchorType[D]] = HashTrieMap()  # type: ignore[reportGeneralTypeIssues]  # noqa: E501
+    _uncrawled: HashTrieSet[URI] = EMPTY_UNCRAWLED
+    _retrieve: Retrieve[D] = field(default=_fail_to_retrieve)
 
     def __getitem__(self, uri: URI) -> Resource[D]:
         """
@@ -252,29 +239,56 @@ class Registry(Mapping[URI, Resource[D]]):
         """
         try:
             return self._resources[uri]
-        except LookupError:
-            try:
-                return self._retrieve(uri)
-            except (
-                exceptions.CannotDetermineSpecification,
-                exceptions.NoSuchResource,
-            ):
-                raise
-            except Exception:
-                raise exceptions.Unretrievable(ref=uri)
+        except KeyError:
+            raise exceptions.NoSuchResource(ref=uri)
 
     def __iter__(self) -> Iterator[URI]:
         """
-        Iterate over all known URIs in the registry.
+        Iterate over all fully crawled URIs in the registry.
         """
         return iter(self._resources)
 
     def __len__(self) -> int:
         """
-        Count the total number of (fully crawled) resources in this registry.
+        Count the total number of fully crawled resources in this registry.
         """
         return len(self._resources)
 
+    def __rmatmul__(self, new: Resource[D] | Iterable[Resource[D]]):
+        """
+        Add resource(s) to a new registry, using their internal IDs.
+
+        Resources must have a internal IDs (e.g. the ``$id`` keyword in modern
+        JSON Schema versions), otherwise an error will be raised.
+
+        Use this via:
+
+            * ``resource @ registry`` or
+
+            * ``[iterable, of, multiple, resources] @ registry``
+
+        which -- again, assuming the resources have internal IDs -- is
+        equivalent to calling `Registry.with_resources` as such:
+
+        .. code:: python
+
+            registry.with_resources(
+                (resource.id(), resource) for resource in new_resources
+            )
+        """
+        if isinstance(new, Resource):
+            new = (new,)
+
+        resources = self._resources
+        uncrawled = self._uncrawled
+        for resource in new:
+            id = resource.id()
+            if id is None:
+                raise exceptions.NoInternalID(resource=resource)
+            uncrawled = uncrawled.insert(id)
+            resources = resources.insert(id, resource)
+        return evolve(self, resources=resources, uncrawled=uncrawled)
+
     def __repr__(self) -> str:
         size = len(self)
         pluralized = "resource" if size == 1 else "resources"
@@ -288,6 +302,32 @@ class Registry(Mapping[URI, Resource[D]]):
             summary = f"{pluralized}"
         return f"<Registry ({size} {summary})>"
 
+    def get_or_retrieve(self, uri: URI):
+        """
+        Get a resource from the registry, crawling or retrieving if necessary.
+        """
+        resource = self._resources.get(uri)
+        if resource is not None:
+            return Retrieved(registry=self, value=resource)
+
+        registry = self.crawl()
+        resource = registry._resources.get(uri)
+        if resource is not None:
+            return Retrieved(registry=registry, value=resource)
+
+        try:
+            resource = registry._retrieve(uri)
+        except (
+            exceptions.CannotDetermineSpecification,
+            exceptions.NoSuchResource,
+        ):
+            raise
+        except Exception:
+            raise exceptions.Unretrievable(ref=uri)
+        else:
+            registry = registry.with_resource(uri, resource)
+            return Retrieved(registry=registry, value=resource)
+
     def remove(self, uri: URI):
         """
         Return a registry with the resource identified by a given URI removed.
@@ -299,7 +339,7 @@ class Registry(Mapping[URI, Resource[D]]):
             self,
             resources=self._resources.remove(uri),
             uncrawled=self._uncrawled.discard(uri),
-            anchors=pmap(
+            anchors=HashTrieMap(
                 (k, v) for k, v in self._anchors.items() if k[0] != uri
             ),
         )
@@ -308,7 +348,15 @@ class Registry(Mapping[URI, Resource[D]]):
         """
         Retrieve the given anchor, which must already have been found.
         """
-        return self._anchors[uri, name]
+        value = self._anchors.get((uri, name))
+        if value is not None:
+            return Retrieved(value=value, registry=self)
+
+        registry = self.crawl()
+        value = registry._anchors.get((uri, name))
+        if value is not None:
+            return Retrieved(value=value, registry=registry)
+        raise exceptions.NoSuchAnchor(ref=uri, resource=self[uri], anchor=name)
 
     def contents(self, uri: URI) -> D:
         """
@@ -320,8 +368,8 @@ class Registry(Mapping[URI, Resource[D]]):
         """
         Immediately crawl all added resources, discovering subresources.
         """
-        resources = self._resources.evolver()
-        anchors = self._anchors.evolver()
+        resources = self._resources
+        anchors = self._anchors
         uncrawled = [(uri, resources[uri]) for uri in self._uncrawled]
         while uncrawled:
             uri, resource = uncrawled.pop()
@@ -329,14 +377,14 @@ class Registry(Mapping[URI, Resource[D]]):
             id = resource.id()
             if id is not None:
                 uri = urljoin(uri, id)
-                resources[uri] = resource
+                resources = resources.insert(uri, resource)
             for each in resource.anchors():
-                anchors.set((uri, each.name), each)
+                anchors = anchors.insert((uri, each.name), each)
             uncrawled.extend((uri, each) for each in resource.subresources())
         return evolve(
             self,
-            resources=resources.persistent(),
-            anchors=anchors.persistent(),
+            resources=resources,
+            anchors=anchors,
             uncrawled=EMPTY_UNCRAWLED,
         )
 
@@ -353,16 +401,12 @@ class Registry(Mapping[URI, Resource[D]]):
         r"""
         Add the given `Resource`\ s to the registry, without crawling them.
         """
-        resources = self._resources.evolver()
-        uncrawled = self._uncrawled.evolver()
+        resources = self._resources
+        uncrawled = self._uncrawled
         for uri, resource in pairs:
-            uncrawled.add(uri)
-            resources[uri] = resource
-        return evolve(
-            self,
-            resources=resources.persistent(),
-            uncrawled=uncrawled.persistent(),
-        )
+            uncrawled = uncrawled.insert(uri)
+            resources = resources.insert(uri, resource)
+        return evolve(self, resources=resources, uncrawled=uncrawled)
 
     def with_contents(
         self,
@@ -392,8 +436,8 @@ class Registry(Mapping[URI, Resource[D]]):
             anchors = anchors.update(registry._anchors)  # type: ignore[reportUnknownMemberType]  # noqa: E501
             uncrawled = uncrawled.update(registry._uncrawled)
 
-            if registry._retrieve != _fail_to_retrieve:
-                if registry._retrieve != retrieve != _fail_to_retrieve:
+            if registry._retrieve is not _fail_to_retrieve:
+                if registry._retrieve is not retrieve is not _fail_to_retrieve:
                     raise ValueError(
                         "Cannot combine registries with conflicting retrieval "
                         "functions.",
@@ -424,10 +468,23 @@ class Registry(Mapping[URI, Resource[D]]):
         )
 
 
+T = TypeVar("T", AnchorType[Any], Resource[Any])
+
+
+@frozen
+class Retrieved(Generic[D, T]):
+    """
+    A value retrieved from a `Registry`.
+    """
+
+    value: T
+    registry: Registry[D]
+
+
 @frozen
 class Resolved(Generic[D]):
     """
-    A resolved reference.
+    A reference resolved to its contents by a `Resolver`.
     """
 
     contents: D
@@ -453,11 +510,7 @@ class Resolver(Generic[D]):
 
     _base_uri: str = field(alias="base_uri")
     _registry: Registry[D] = field(alias="registry")
-    _previous: PList[URI] = field(
-        default=EMPTY_PREVIOUS_RESOLVERS,
-        repr=False,
-        alias="previous",
-    )
+    _previous: List[URI] = field(default=List(), repr=False, alias="previous")
 
     def lookup(self, ref: URI) -> Resolved[D]:
         """
@@ -486,44 +539,24 @@ class Resolver(Generic[D]):
             uri, fragment = self._base_uri, ref[1:]
         else:
             uri, fragment = urldefrag(urljoin(self._base_uri, ref))
-        registry = self._registry
-        resource = registry.get(uri)
-        if resource is None:
-            registry = registry.crawl()
-            try:
-                resource = registry[uri]
-            except exceptions.NoSuchResource:
-                raise exceptions.Unresolvable(ref=ref) from None
-            except exceptions.Unretrievable:
-                raise exceptions.Unresolvable(ref=ref)
+        try:
+            retrieved = self._registry.get_or_retrieve(uri)
+        except exceptions.NoSuchResource:
+            raise exceptions.Unresolvable(ref=ref) from None
+        except exceptions.Unretrievable:
+            raise exceptions.Unresolvable(ref=ref)
 
         if fragment.startswith("/"):
-            return resource.pointer(
-                pointer=fragment,
-                resolver=self._evolve(registry=registry, base_uri=uri),
-            )
+            resolver = self._evolve(registry=retrieved.registry, base_uri=uri)
+            return retrieved.value.pointer(pointer=fragment, resolver=resolver)
 
         if fragment:
-            try:
-                anchor = registry.anchor(uri, fragment)
-            except LookupError:
-                registry = registry.crawl()
-                try:
-                    anchor = registry.anchor(uri, fragment)
-                except LookupError:
-                    raise exceptions.NoSuchAnchor(
-                        ref=ref,
-                        resource=resource,
-                        anchor=fragment,
-                    )
-            return anchor.resolve(
-                resolver=self._evolve(registry=registry, base_uri=uri),
-            )
+            retrieved = retrieved.registry.anchor(uri, fragment)
+            resolver = self._evolve(registry=retrieved.registry, base_uri=uri)
+            return retrieved.value.resolve(resolver=resolver)
 
-        return Resolved(
-            contents=resource.contents,
-            resolver=self._evolve(registry=registry, base_uri=uri),
-        )
+        resolver = self._evolve(registry=retrieved.registry, base_uri=uri)
+        return Resolved(contents=retrieved.value.contents, resolver=resolver)
 
     def in_subresource(self, subresource: Resource[D]) -> Resolver[D]:
         """
@@ -547,7 +580,7 @@ class Resolver(Generic[D]):
         """
         previous = self._previous
         if self._base_uri and (not previous or base_uri != self._base_uri):
-            previous = previous.cons(self._base_uri)
+            previous = previous.push_front(self._base_uri)
         return evolve(self, base_uri=base_uri, previous=previous, **kwargs)
 
 
